// index.js (Backend)

const express = require('express');
const cors = require('cors');
const mongoose = require('mongoose');
require('dotenv').config(); // بارگذاری متغیرهای محیطی از .env

// 1️⃣ اتصال به MongoDB
const MONGO_URI = process.env.MONGO_URI;
mongoose
  .connect(MONGO_URI)
  .then(() => console.log('✅ MongoDB connected!'))
  .catch(err => {
    console.error('❌ MongoDB connection error:', err.message);
    process.exit(1);
  });

// 2️⃣ تعریف اسکیمای Patient
const patientSchema = new mongoose.Schema(
  {
    fullName: { type: String, required: true },
    birthDate: Date,
    phone: { type: String, required: true },
    address: String,
    tag: String,
    photos: [String], // URLs or filenames
  },
  { timestamps: true }
);
const Patient = mongoose.model('Patient', patientSchema);

// 3️⃣ تعریف اسکیمای Appointment
const appointmentSchema = new mongoose.Schema(
  {
    patientId: { type: mongoose.Schema.Types.ObjectId, ref: 'Patient', required: true },
    dateShamsi: { type: String, required: true },
    time: String,
    type: { type: String, enum: ['Injection', 'Laser'], default: 'Injection' },
    status: { type: String, enum: ['Scheduled','Completed','Canceled'], default: 'Scheduled' },
  },
  { timestamps: true }
);
const Appointment = mongoose.model('Appointment', appointmentSchema);

// 4️⃣ راه‌اندازی Express و Middlewareها
const app = express();
app.use(cors());            // فعال‌سازی CORS برای همه‌ی روت‌ها
app.use(express.json());    // پارس کردن JSON body درخواست‌ها

// 5️⃣ روت‌های CRUD بیماران

// لیست تمام بیماران
app.get('/patients', async (req, res) => {
  const patients = await Patient.find().sort({ createdAt: -1 });
  res.json(patients);
});

// ساخت بیمار جدید
app.post('/patients', async (req, res) => {
  try {
    const patient = new Patient(req.body);
    const saved = await patient.save();
    res.status(201).json(saved);
  } catch (err) {
    res.status(400).json({ error: err.message });
  }
});

// ویرایش بیمار
app.put('/patients/:id', async (req, res) => {
  try {
    const updated = await Patient.findByIdAndUpdate(req.params.id, req.body, { new: true });
    res.json(updated);
  } catch (err) {
    res.status(400).json({ error: err.message });
  }
});

// حذف بیمار
app.delete('/patients/:id', async (req, res) => {
  await Patient.findByIdAndDelete(req.params.id);
  res.status(204).end();
});

// 6️⃣ روت‌های CRUD نوبت‌ها

// لیست نوبت‌ها (اختیاری با فیلتر تاریخ شمسی)
app.get('/appointments', async (req, res) => {
  const { dateShamsi } = req.query;
  const filter = dateShamsi ? { dateShamsi } : {};
  const list = await Appointment.find(filter)
    .populate('patientId', 'fullName phone')
    .sort({ dateShamsi: 1, time: 1 });
  res.json(list);
});

// ساخت نوبت جدید
app.post('/appointments', async (req, res) => {
  try {
    const appt = new Appointment(req.body);
    const saved = await appt.save();
    const populated = await saved.populate('patientId', 'fullName phone');
    res.status(201).json(populated);
  } catch (err) {
    res.status(400).json({ error: err.message });
  }
});

// ویرایش نوبت
app.put('/appointments/:id', async (req, res) => {
  try {
    const updated = await Appointment.findByIdAndUpdate(req.params.id, req.body, { new: true })
      .populate('patientId', 'fullName phone');
    res.json(updated);
  } catch (err) {
    res.status(400).json({ error: err.message });
  }
});

// حذف نوبت
app.delete('/appointments/:id', async (req, res) => {
  await Appointment.findByIdAndDelete(req.params.id);
  res.status(204).end();
});

// 7️⃣ استارت کردن سرور
const PORT = process.env.PORT || 5000;
app.listen(PORT, () => {
  console.log(`🚀 Server running on http://localhost:${PORT}`);
});
